#!/bin/sh
# FILE: oommf.tcl
#
# This Tcl script is the master application of the OOMMF project.
# It must be evaluted by a Tcl interpreter such as tclsh or wish
# (or whatever other name they are known by on your platform).
#
# On Unix platforms, it may be necessary to edit line 17 of this file
# so that the second word is the name of a Tcl interpreter shell
# installed on your computer.
#
# On Windows platforms, when Tcl/Tk was installed, an association was
# placed in the registry to recognize the file extension .tcl of this
# file and use the installed wish program to evaluate it.
#
#    v--Edit here if necessary \
exec tclsh "$0" ${1+"$@"}
#
set oommf(version) 1.1.0.0
set oommf(defaultArgv) mmLaunch
set oommf(usage) [subst {
Usage:
-----
oommf.tcl ?options? ?app-spec? ?args?
        where
Valid options:
        -fg     Run app in foreground; block till it completes
        -bg     Run app in background; exit after starting it (default)
        -command        Print app's command line to stdout and quit
        -version        Print version of oommf.tcl and quit
        -platform       Print platform name and quit
        -h*     Print usage summary and quit
        --      End of options, next argument is app-spec
Valid app-spec: 'appName ?-exact? ?version-requirement?'
        (NB: app-spec is quoted as a single argument)
        (NB2: Default is '$oommf(defaultArgv)'.)
        where
                appName                 Name of app to start (case insensitive)
                -exact                  When present, version-requirement
                                        is exact; later releases are not
                                        acceptable
                version-requirement     Minimum (or exact) version which
                                        is acceptable.  When omitted,
                                        accept latest version available.
?args?
        rest of command line is passed as command line arguments
        of the application started.
}]
#
########################################################################
# Make sure none of the boiler plate code consumes parts of the
# command line meant for our child processes.
set saveArgv $argv
set argv ""

########################################################################
# This script is meant to be portable to many platforms and many
# versions of Tcl/Tk.  Reporting of errors in startup scripts wasn't
# reliable on all platforms until Tcl version 8.0, so this script
# catches it's own errors and reports them to support older Tcl
# releases.
#
# The following [catch] command catches any errors which arise in the
# main body of this script.
#
set code [catch {

if {[catch {package require Tcl 8}]} {
    if {[catch {package require Tcl 7.5}]} {
	return -code error "OOMMF software requires Tcl 7.5 or\
		higher.\n\tYour version: Tcl [info patchlevel]\nPlease\
		upgrade."
    }
}

# When this script is evaluated by a shell which doesn't include the
# Oc extension, we must find and load the Oc extension on our own.
if {[string match "" [package provide Oc]]} {

# Robust determination of the absolute, direct pathname of this script:
set cwd [pwd]
set myAbsoluteName [file join $cwd [info script]]
if {[catch {cd [file dirname $myAbsoluteName]} msg]} {
    # This is probably impossible.  Can we read a script if we can't
    # cd to the directory which contains it?
    return -code error "Can't determine direct pathname for\
	    $myAbsoluteName:\n\t$msg"
}
set myAbsoluteName [file join [pwd] [file tail $myAbsoluteName]]
cd $cwd

# Create search path for OOMMF runtime library
set oommf(libPath) {}
if {[info exists env(OOMMF_ROOT)] \
        && [string match absolute [file pathtype $env(OOMMF_ROOT)]] \
        && [file isdirectory [file join $env(OOMMF_ROOT) app]]} {
    lappend oommf(libPath) $env(OOMMF_ROOT)
}
set _ [file dirname $myAbsoluteName]
if {[file isdirectory [file join $_ app]]} {
    lappend oommf(libPath) $_
}
if {![llength $oommf(libPath)]} {
    return -code error "Unable to find your OOMMF installation.\n\nSet the\
            environment variable OOMMF_ROOT to its root directory."
}

# Construct search path for extensions
foreach lib $oommf(libPath) {
    set dir [file join $lib ext]
    if {[file isdirectory $dir] && [file readable $dir] \
            && [string match absolute [file pathtype $dir]]} {
        if {[lsearch -exact $auto_path $dir] < 0} {
            lappend auto_path $dir
        }
    }
}

# Locate a version of the Oc (OOMMF core) extension and set it up for loading.
if {[catch {package require Oc} msg]} {
    # When can 'package require Oc' raise an error?
    #	1. There's no Oc extension installed.
    #	2. An error occurred in the initialization of the Oc extension.
    #	3. Some other extension has a broken pkgIndex.tcl file, and
    #	   we are running a pre-8.0 version of Tcl, which didn't
    #	   control such problems.
    #
    # In case 3 only, it makes sense to try again.  Why?  Because the
    # pkgIndex.tcl file of the Oc extension may have been sourced before
    # the error.  Then on the second try, the [package unknown] step of
    # [package require] will be bypassed, so the error won't occur again.
    #
    # Otherwise, we should just report the error.
    if {[llength [package versions Oc]]
	    && [string match "" [package provide Oc]]} {
	package require Oc
    } else {
	error $msg $errorInfo $errorCode
    }
}
# Cleanup globals created by boilerplate code
foreach varName {cwd myAbsoluteName msg _ lib dir} {
    catch {unset $varName}
}
unset varName

# End of code for finding/loading the Oc extension
}

########################################################################
########################################################################
#
# Here's where the non-boilerplate code starts
#
########################################################################
########################################################################

# Restore command line
set argv $saveArgv
unset saveArgv

if {[Oc_Main HasTk]} {
    wm withdraw .
}

# Process the options
set bg 1
while {[string match -* [lindex $argv 0]]} {
    regexp -- -(.*) [lindex $argv 0] _ option
    set argv [lrange $argv 1 end]
    switch -exact -- $option {
        - {
            break
        }
        version {
	    # Should this write directly to stdout ?
	    #	PRO: Perhaps that's where output is expected?
	    #	CON: not as portable
            Oc_Log Log "\nOOMMF bootstrap $oommf(version)" info
            exit
        }
        platform {
            Oc_Log Log "\nPlatform name:\t[[Oc_Config RunPlatform] \
                    GetValue platform_name]" info
            exit
        }
        command {
            while {[string match -* [lindex $argv 0]]} {
                regexp -- -(.*) [lindex $argv 0] _ option
                set argv [lrange $argv 1 end]
                if {[string match - $option]} {break}
            }
            set code [catch {eval Oc_Application CommandLine $argv} msg]
            if {$code} {
                Oc_Log Log $msg panic
            } else {
		# Should this use Oc_Log?  
		#	PRO: more cross-platform?
		#	CON: usually means writing to stderr 
                puts stdout $msg
            }
            exit $code
        }
        fg {
            set bg 0
        }
        bg {
            set bg 1
        }
        default {
            if {[string match h* $option]} {
                Oc_Log Log $oommf(usage) info
                exit
            }
            return -code error "Unrecognized option: -$option\n$oommf(usage)"
        }
    }
}
# Use default if no command line
if {[llength $argv] == 0} {
    set argv $oommf(defaultArgv)
}

if {$bg} {
    if {[catch {eval Oc_Application Exec $argv &} msg]} {
        # Couldn't launch background app; report the problem
        return -code error $msg
    }
} else {
    if {[catch {eval Oc_Application CommandLine $argv} cmd]} {
        return -code error $cmd
    }
    if {[catch {eval Oc_Exec Foreground $cmd} msg]} {
	if {[string compare "child process exited abnormally: " $msg] == 0} {
	    # Expected error message - don't bother displaying
	    exit 1
	}
	return -code error $msg
    }
}

########################################################################
########################################################################
#
# End of the non-boilerplate code.
#
########################################################################
########################################################################

# The following line is the end of the [catch] command which encloses the
# entire main body of this script.
} msg]
#
########################################################################

########################################################################
# Check the result of the enclosing [catch] and construct any error
# message to be reported.
#
if {$code == 0} {
    # No error; we're done
    exit 0
} elseif {$code == 1} {
    # Unexpected error; Report whole stack trace plus errorCode
    set rei $errorInfo
    set rec $errorCode
} elseif {$code != 2} {
    # Weirdness -- just pass it on
    return -code $code $msg
} else {
    # Expected error; just report the message
    set rei $msg
    set rec $errorCode
}
#
########################################################################

########################################################################
# Try various error reporting methods
#
# First, use Oc_Log if it is available.
set src ""
if {[string match OC [lindex $rec 0]]} {
    set src [lindex $rec 1]
}
set errorInfo $rei
set errorCode $rec
if {![catch {Oc_Log Log $msg panic $src}]} {
    # Oc_Log reported the error; now die
    exit 1
}

# If we're on Windows using pre-8.0 Tk, we have to report the message
# ourselves.  Use [tk_dialog].
if {[string length [package provide Tk]]
        && [package vsatisfies [package provide Tk] 4]
        && [array exists tcl_platform]
        && [string match windows $tcl_platform(platform)]} {
    wm withdraw .
    option add *Dialog.msg.wrapLength 0
    tk_dialog .error "<[pid]> [info script] error:" $rei error 0 Die
    exit 1
}

# Fall back on Tcl/Tk's own reporting of startup script errors
error $msg "\n<[pid]> [info script] error:\n$rei" $rec
#
########################################################################
